
/*
** Copyright (c) 2006 Herbert Poetzl
**
** V0.01	initial prototype
**
*/

#define VERSION "V0.01"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>


#include "vserver.h"
#include "switch.h"

#include "context_cmd.h"
#include "debug_cmd.h"
#include "inode_cmd.h"
#include "namespace_cmd.h"
#include "sched_cmd.h"
#include "cvirt_cmd.h"
#include "dlimit_cmd.h"
#include "limit_cmd.h"
#include "network_cmd.h"
#include "signal_cmd.h"


static	char	opt_id = 0;
static	char	opt_cmd = 0;
static	char	opt_vers = 0;
static	char	opt_desc = 0;
static	char	opt_cont = 0;
static	char	opt_list = 0;
static	char	opt_quiet = 0;
static	char	opt_terse = 0;
static	char	opt_after = 0;
static	char	opt_before = 0;

static	char *	cmd_name = NULL;
static	char *	exec_name = NULL;

static uint32_t	num_id = 0;
static uint32_t	num_ret = 0;
static uint32_t	num_cmd = 0;

static int	num_err = 0;


long long int argtoll(const char *str)
{
	int bit, inv = 0;
	long long int val = 0;
	char *eptr;
	
	switch (*str) {
	case '~':
	case '!':
	    inv = 1;	/* invert */
	    str++;
	default:
	    break;
	}

	while (*str) {
	    switch (*str) {
	    case '^':
		bit = strtol(str+1, &eptr, 0);
		val ^= (1LL << bit);
		break;
	    case '&':
		val &= strtoll(str+1, &eptr, 0);
		break;
	    case '|':
		val |= strtoll(str+1, &eptr, 0);
		break;
	    case '.':
		val = (val << 8) | (strtoll(str+1, &eptr, 0) & 0xff);
		break;
	    default:
		val = strtoll(str, &eptr, 0);
		break;
	    }
	    if (eptr == str)
	    	break;
	    str = eptr;
	}
	return (inv)?~(val):(val);
}


#define	ATO_TYPE(_type)			\
_type ato_##_type (const char *str) {	\
	_type val = argtoll(str);	\
	return val;			\
}

ATO_TYPE(uint64_t)
#define	FMT_uint64_t	"0x%016llX (%llu)"

ATO_TYPE(uint32_t)
#define	FMT_uint32_t	"0x%08llX (%llu)"

ATO_TYPE(uint16_t)
#define	FMT_uint16_t	"0x%04llX (%llu)"

ATO_TYPE(int64_t)
#define	FMT_int64_t	"0x%016llX (%lld)"

ATO_TYPE(int32_t)
#define	FMT_int32_t	"0x%08llX (%lld)"

ATO_TYPE(int16_t)
#define	FMT_int16_t	"0x%04llX (%lld)"


#define	FMT_INDEX	"  [%d] .%-16s"


#define	CMD_COND(_cmd,_index,_out)			\
	((out == _out) && (cmd == _cmd) && (index == _index))


#define	CMD_FILL_T(_cmd,_index,_type,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,0)) {			\
	    data->_ent._name = ato_##_type(str);	\
	    return 1;					\
	}} while (0)

#define	CMD_FILL_S(_cmd,_index,_len,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,0)) {			\
	    strncpy(data->_ent._name, str, _len);	\
	    return 1;					\
	}} while (0)

#define	CMD_FILL_P(_cmd,_index,_ent,_name) do {		\
	if (CMD_COND(_cmd,_index,0)) {			\
	    data->_ent._name = str;			\
	    return 1;					\
	}} while (0)


#define	CMD_OUT_T(_cmd,_index,_type,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,1)) {			\
	    printf(FMT_INDEX " " FMT_##_type "\n",	\
		_index, #_name,				\
		(long long)data->_ent._name,		\
		(long long)data->_ent._name);		\
	    return 1;					\
	}} while (0)

#define	CMD_OUT_S(_cmd,_index,_len,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,1)) {			\
	    printf(FMT_INDEX " %.*s\n", _index,	\
		#_name, _len, data->_ent._name);	\
	    return 1;					\
	}} while (0)

#define	CMD_OUT_P(_cmd,_index,_ent,_name) do {		\
	if (CMD_COND(_cmd,_index,1)) {			\
	    printf(FMT_INDEX " PTR: %p %s\n",		\
		_index, #_name,				\
		data->_ent._name,			\
		data->_ent._name);		\
	    return 1;					\
	}} while (0)


#define	CMD_ICOND(_cmd) ((out == -1) && (cmd == _cmd))

#define	CMD_INDEX(_cmd,_index,_name) do {		\
	if (CMD_ICOND(_cmd) && (strcmp(str,_name)==0))	\
	    return _index;				\
	} while (0)

#define	CMD_VCMD(_cmd,_name,_index,_io) do {		\
	if ((out == -2) && (strcmp(str,_name)==0))	\
	    return _cmd;				\
	if (CMD_COND(_cmd,_index,-2))			\
	    return _io;					\
	} while (0)

#define	CMD_DESC(_cmd,_name,_index,_io,_type) do {	\
	if (CMD_COND(_cmd,_index,-3)) {			\
	    printf(FMT_INDEX " (%s)\n",			\
		_index, _name,				\
		(_io == 3)?"in/out":			\
		((_io == 1)?"in":"out"));		\
	    return 1;					\
	}} while (0)


#define	CMDE_T(_cmd,_ext,_index,_io,_type,_name) do {		\
	CMD_OUT_T(VCMD_##_cmd,_index,_type,_cmd##_ext,_name);	\
	CMD_FILL_T(VCMD_##_cmd,_index,_type,_cmd##_ext,_name);	\
	CMD_INDEX(VCMD_##_cmd,_index,#_name);			\
	CMD_DESC(VCMD_##_cmd,#_name,_index,_io,#_type);		\
	CMD_VCMD(VCMD_##_cmd,#_cmd,_index,_io);			\
	} while (0)

#define	CMDE_S(_cmd,_ext,_index,_io,_len,_name) do {		\
	CMD_OUT_S(VCMD_##_cmd,_index,_len,_cmd##_ext,_name);	\
	CMD_FILL_S(VCMD_##_cmd,_index,_len,_cmd##_ext,_name);	\
	CMD_INDEX(VCMD_##_cmd,_index,#_name);			\
	CMD_DESC(VCMD_##_cmd,#_name,_index,_io,"str_"#_len);	\
	CMD_VCMD(VCMD_##_cmd,#_cmd,_index,_io);			\
	} while (0)

#define	CMDE_P(_cmd,_ext,_index,_io,_name) do {			\
	CMD_OUT_P(VCMD_##_cmd,_index,_cmd##_ext,_name);		\
	CMD_FILL_P(VCMD_##_cmd,_index,_cmd##_ext,_name);	\
	CMD_INDEX(VCMD_##_cmd,_index, #_name);			\
	CMD_DESC(VCMD_##_cmd,#_name,_index,_io,"ptr");		\
	CMD_VCMD(VCMD_##_cmd,#_cmd,_index,_io);			\
	} while (0)

#define	CMDE_X(_cmd) do {					\
	CMD_VCMD(VCMD_##_cmd,#_cmd,-1,0);			\
	} while (0)


#define	EENTRY(_symb,_desc) {	\
	.code = _symb,		\
	.symb = #_symb,		\
	.desc = _desc,		\
	}


struct	err_data {
	int code;
	char *symb;
	char *desc;
} err_data[] = {
	EENTRY(EPERM,	"Insufficient Priviledges"),
	EENTRY(ENOENT,	"No Such Entity"),
	EENTRY(ESRCH,	"No Such Process or Context"),
	EENTRY(EINTR,	"Interrupted Syscall"),
	EENTRY(ENXIO,	"No Such Device or Address"),
	EENTRY(EAGAIN,	"Try Again"),
	EENTRY(ENOMEM,	"Out Of Memeory"),
	EENTRY(EACCES,	"Permission Denied"),
	EENTRY(EFAULT,	"Bad Address"),
	EENTRY(ENOTTY,	"Not A Typewriter :)"),
	EENTRY(ENOSYS,	"Function Not Implemented"),
	EENTRY(0,	"Unknown Error"),
};


union	cmd_arg {

	/* from context_cmd */
	struct vcmd_vx_info_v0		vx_info;
	struct vcmd_ctx_create		ctx_create;
	struct vcmd_ctx_migrate		ctx_migrate;
	struct vcmd_ctx_flags_v0	get_cflags;
	struct vcmd_ctx_flags_v0	set_cflags;
	struct vcmd_ctx_caps_v0		get_ccaps;
	struct vcmd_ctx_caps_v0		set_ccaps;

	/* from network_cmd */
	struct vcmd_nx_info_v0		nx_info;
	struct vcmd_net_create		net_create;
	struct vcmd_net_addr_v0		net_add;
	struct vcmd_net_addr_v0		net_remove;
	struct vcmd_net_flags_v0	get_nflags;
	struct vcmd_net_flags_v0	set_nflags;
	struct vcmd_net_caps_v0		get_ncaps;
	struct vcmd_net_caps_v0		set_ncaps;

	/* from cvirt_cmd */
	struct vcmd_vhi_name_v0		set_vhi_name;
	struct vcmd_vhi_name_v0		get_vhi_name;

	/* from dlimit_cmd */
	struct vcmd_ctx_dlimit_base_v0	add_dlimit;
	struct vcmd_ctx_dlimit_base_v0	rem_dlimit;
	struct vcmd_ctx_dlimit_v0	set_dlimit;
	struct vcmd_ctx_dlimit_v0	get_dlimit;

	/* from inode_cmd */
	struct vcmd_ctx_iattr_v0	get_iattr_v0;
	struct vcmd_ctx_iattr_v0	set_iattr_v0;
	struct vcmd_ctx_iattr_v1	get_iattr_v1;
	struct vcmd_ctx_iattr_v1	set_iattr_v1;

	/* from rlimit_cmd */
	struct vcmd_ctx_rlimit_v0	get_rlimit;
	struct vcmd_ctx_rlimit_v0	set_rlimit;
	struct vcmd_ctx_rlimit_mask_v0	get_rlimit_mask;

	/* from sched_cmd */
	struct vcmd_set_sched_v2	set_sched_v2;
	struct vcmd_set_sched_v3	set_sched_v3;
	struct vcmd_set_sched_v4	set_sched_v4;

	/* from signal_cmd */
	struct vcmd_ctx_kill_v0		ctx_kill;
	struct vcmd_wait_exit_v0	wait_exit;
};



int	handle_data(union cmd_arg *data, const char *str,
	uint32_t cmd, int index, int out)
{
#ifdef	DEBUG
	printf(" hadle_data(%p,%s,%x,%d,%d)\n",
	    data, str, cmd, index, out);
#endif
	CMDE_X(get_version		);
	CMDE_X(get_vci			);

	/* from context_cmd */

	CMDE_T(vx_info,,		0,2, uint32_t,	xid);
	CMDE_T(vx_info,,		1,2, uint32_t,	initpid);

	CMDE_X(ctx_create_v0		);
	CMDE_T(ctx_create,,		0,1, uint64_t,	flagword);

	CMDE_X(ctx_create_v0		);
	CMDE_T(ctx_migrate,,		0,1, uint64_t,	flagword);

	CMDE_T(get_cflags,,		0,2, uint64_t,	flagword);
	CMDE_T(get_cflags,,		1,2, uint64_t,	mask);

	CMDE_T(set_cflags,,		0,3, uint64_t,	flagword);
	CMDE_T(set_cflags,,		1,3, uint64_t,	mask);

	CMDE_T(get_ccaps,,		0,2, uint64_t,	bcaps);
	CMDE_T(get_ccaps,,		1,2, uint64_t,	ccaps);
	CMDE_T(get_ccaps,,		2,2, uint64_t,	cmask);

	CMDE_T(set_ccaps,,		0,1, uint64_t,	bcaps);
	CMDE_T(set_ccaps,,		1,1, uint64_t,	ccaps);
	CMDE_T(set_ccaps,,		2,1, uint64_t,	cmask);


	/* from network_cmd */

	CMDE_T(nx_info,,		0,2, uint32_t,	nid);

	CMDE_X(net_create_v0		);
	CMDE_T(net_create,,		0,1, uint64_t,	flagword);

	CMDE_X(net_migrate		);

	CMDE_T(net_add,,		0,1, uint16_t,	type);
	CMDE_T(net_add,,		1,1, uint16_t,	count);
	CMDE_T(net_add,,		2,1, uint32_t,	ip[0]);
	CMDE_T(net_add,,		3,1, uint32_t,	mask[0]);
	CMDE_T(net_add,,		4,1, uint32_t,	ip[1]);
	CMDE_T(net_add,,		5,1, uint32_t,	mask[1]);
	CMDE_T(net_add,,		6,1, uint32_t,	ip[2]);
	CMDE_T(net_add,,		7,1, uint32_t,	mask[2]);
	CMDE_T(net_add,,		8,1, uint32_t,	ip[3]);
	CMDE_T(net_add,,		9,1, uint32_t,	mask[3]);

	CMDE_T(net_remove,,		0,1, uint16_t,	type);
	CMDE_T(net_remove,,		1,1, uint16_t,	count);
	CMDE_T(net_remove,,		2,1, uint32_t,	ip[0]);
	CMDE_T(net_remove,,		3,1, uint32_t,	mask[0]);
	CMDE_T(net_remove,,		4,1, uint32_t,	ip[1]);
	CMDE_T(net_remove,,		5,1, uint32_t,	mask[1]);
	CMDE_T(net_remove,,		6,1, uint32_t,	ip[2]);
	CMDE_T(net_remove,,		7,1, uint32_t,	mask[2]);
	CMDE_T(net_remove,,		8,1, uint32_t,	ip[3]);
	CMDE_T(net_remove,,		9,1, uint32_t,	mask[3]);

	CMDE_T(get_nflags,,		0,2, uint64_t,	flagword);
	CMDE_T(get_nflags,,		1,2, uint64_t,	mask);

	CMDE_T(set_nflags,,		0,1, uint64_t,	flagword);
	CMDE_T(set_nflags,,		1,1, uint64_t,	mask);

	CMDE_T(get_ncaps,,		0,2, uint64_t,	ncaps);
	CMDE_T(get_ncaps,,		1,2, uint64_t,	cmask);

	CMDE_T(set_ncaps,,		0,1, uint64_t,	ncaps);
	CMDE_T(set_ncaps,,		1,1, uint64_t,	cmask);


	/* from cvirt_cmd */

	CMDE_T(set_vhi_name,,		0,1, uint32_t,	field);
	CMDE_S(set_vhi_name,,		1,1, 65,	name);

	CMDE_T(get_vhi_name,,		0,3, uint32_t,	field);
	CMDE_S(get_vhi_name,,		1,2, 65,	name);


	/* from dlimit_cmd */

	CMDE_P(add_dlimit,,		0,1,		name);
	CMDE_T(add_dlimit,,		1,1, uint32_t,	flags);

	CMDE_P(rem_dlimit,,		0,1,		name);
	CMDE_T(rem_dlimit,,		1,1, uint32_t,	flags);

	CMDE_P(set_dlimit,,		0,1,		name);
	CMDE_T(set_dlimit,,		1,1, uint32_t,	space_used);
	CMDE_T(set_dlimit,,		2,1, uint32_t,	space_total);
	CMDE_T(set_dlimit,,		3,1, uint32_t,	inodes_used);
	CMDE_T(set_dlimit,,		4,1, uint32_t,	inodes_total);
	CMDE_T(set_dlimit,,		5,1, uint32_t,	reserved);
	CMDE_T(set_dlimit,,		6,1, uint32_t,	flags);

	CMDE_P(get_dlimit,,		0,1,		name);
	CMDE_T(get_dlimit,,		1,2, uint32_t,	space_used);
	CMDE_T(get_dlimit,,		2,2, uint32_t,	space_total);
	CMDE_T(get_dlimit,,		3,2, uint32_t,	inodes_used);
	CMDE_T(get_dlimit,,		4,2, uint32_t,	inodes_total);
	CMDE_T(get_dlimit,,		5,2, uint32_t,	reserved);
	CMDE_T(get_dlimit,,		6,2, uint32_t,	flags);


	/* from inode_cmd */

	CMDE_T(get_iattr_v0,,		0,1, uint64_t,	ino);
	CMDE_T(get_iattr_v0,,		1,3, uint32_t,	xid);
	CMDE_T(get_iattr_v0,,		2,2, uint32_t,	flags);
	CMDE_T(get_iattr_v0,,		3,2, uint32_t,	mask);

	CMDE_T(set_iattr_v0,,		0,1, uint64_t,	ino);
	CMDE_T(set_iattr_v0,,		1,3, uint32_t,	xid);
	CMDE_T(set_iattr_v0,,		2,3, uint32_t,	flags);
	CMDE_T(set_iattr_v0,,		3,3, uint32_t,	mask);

	CMDE_P(get_iattr,_v1,		0,1,		name);
	CMDE_T(get_iattr,_v1,		1,3, uint32_t,	xid);
	CMDE_T(get_iattr,_v1,		2,2, uint32_t,	flags);
	CMDE_T(get_iattr,_v1,		3,2, uint32_t,	mask);

	CMDE_P(set_iattr,_v1,		0,1,		name);
	CMDE_T(set_iattr,_v1,		1,3, uint32_t,	xid);
	CMDE_T(set_iattr,_v1,		2,3, uint32_t,	flags);
	CMDE_T(set_iattr,_v1,		3,3, uint32_t,	mask);


	/* from rlimit_cmd */

	CMDE_T(get_rlimit,,		0,1, uint32_t,	id);
	CMDE_T(get_rlimit,,		1,2, uint64_t,	minimum);
	CMDE_T(get_rlimit,,		2,2, uint64_t,	softlimit);
	CMDE_T(get_rlimit,,		3,2, uint64_t,	maximum);

	CMDE_T(set_rlimit,,		0,1, uint32_t,	id);
	CMDE_T(set_rlimit,,		1,1, uint64_t,	minimum);
	CMDE_T(set_rlimit,,		2,1, uint64_t,	softlimit);
	CMDE_T(set_rlimit,,		3,1, uint64_t,	maximum);

	CMDE_T(get_rlimit_mask,,	0,2, uint32_t,	minimum);
	CMDE_T(get_rlimit_mask,,	1,2, uint32_t,	softlimit);
	CMDE_T(get_rlimit_mask,,	2,2, uint32_t,	maximum);


	/* from sched_cmd */

	CMDE_T(set_sched_v2,,		0,1, int32_t,	fill_rate);
	CMDE_T(set_sched_v2,,		1,1, int32_t,	interval);
	CMDE_T(set_sched_v2,,		2,1, int32_t,	tokens);
	CMDE_T(set_sched_v2,,		3,1, int32_t,	tokens_min);
	CMDE_T(set_sched_v2,,		4,1, int32_t,	tokens_max);
	CMDE_T(set_sched_v2,,		5,1, uint64_t,	cpu_mask);

	CMDE_T(set_sched_v3,,		0,1, uint32_t,	set_mask);
	CMDE_T(set_sched_v3,,		1,1, int32_t,	fill_rate);
	CMDE_T(set_sched_v3,,		2,1, int32_t,	interval);
	CMDE_T(set_sched_v3,,		3,1, int32_t,	tokens);
	CMDE_T(set_sched_v3,,		4,1, int32_t,	tokens_min);
	CMDE_T(set_sched_v3,,		5,1, int32_t,	tokens_max);
	CMDE_T(set_sched_v3,,		6,1, int32_t,	priority_bias);

	CMDE_T(set_sched,_v4,		0,1, uint32_t,	set_mask);
	CMDE_T(set_sched,_v4,		1,1, int32_t,	fill_rate);
	CMDE_T(set_sched,_v4,		2,1, int32_t,	interval);
	CMDE_T(set_sched,_v4,		3,1, int32_t,	tokens);
	CMDE_T(set_sched,_v4,		4,1, int32_t,	tokens_min);
	CMDE_T(set_sched,_v4,		5,1, int32_t,	tokens_max);
	CMDE_T(set_sched,_v4,		6,1, int32_t,	prio_bias);
	CMDE_T(set_sched,_v4,		7,1, int32_t,	cpu_id);
	CMDE_T(set_sched,_v4,		8,1, int32_t,	bucket_id);


	/* from signal_cmd */

	CMDE_T(ctx_kill,,		0,1, int32_t,	pid);
	CMDE_T(ctx_kill,,		1,1, int32_t,	sig);

	CMDE_T(wait_exit,,		0,2, int32_t,	reboot_cmd);
	CMDE_T(wait_exit,,		1,2, int32_t,	exit_code);

#ifdef	DEBUG
	printf(" hadle_data(%p,%s,%x,%d,%d) ret=-1\n",
	    data, str, cmd, index, out);
#endif

	return -1;
}


int	set_index(union cmd_arg *data, uint32_t cmd, int index, char *str)
{
	int ret;
#ifdef	DEBUG
	printf(" set index %d = %s\n", index, str);
#endif
	ret = handle_data(data, str, cmd, index, 0);
	if (ret == -1)
	    return -1;
	return ++index;
}

int	set_named(union cmd_arg *data, uint32_t cmd, char *str)
{
	char *pos = index(str, '=');
	int index;

	if (!pos)
	    return -1;

	*pos = '\0';
	index = handle_data(data, str, cmd, 0, -1);
	*pos = '=';
	if (index == -1)
	    return -1; 
#ifdef	DEBUG
	printf(" index of %s = %d\n", str, index);
#endif
	return set_index(data, cmd, index, pos+1);
}

int	parse_cmd(char *str)
{
	int cmd = 0;

	if (strncmp(str, "VCMD_", 5) == 0)
	    str += 5;
	cmd = handle_data(NULL, str, 0, -2, -2);
#ifdef	DEBUG
	printf(" cmd from %s = %x\n", str, cmd);
#endif
	return cmd;
}

int	get_io(uint32_t cmd, int index)
{
	return handle_data(NULL, "", cmd, index, -2);
}

int	has_data(uint32_t cmd)
{
	return handle_data(NULL, "", cmd, -1, -2);
}


#define	CMD_FMT		"VCMD_%02x_%02x[%d]"
#define	CMD_ARG(c)	VC_CATEGORY(c), VC_COMMAND(c), VC_VERSION(c)


void	dump_data(union cmd_arg *data, int out)
{
	int index = 0;
	int io;

	printf(CMD_FMT " {\n", CMD_ARG(num_cmd));
	index = 0;
	do {
	    int r = 0;

	    io = get_io(num_cmd, index);
	    if (opt_terse &&
		((io & out) == 0))
		goto skip;

	    r = handle_data(data, NULL,
		num_cmd, index, 1);
	    if (r < 0)
		break;
	skip:
	    index++;
	} while (io);
	printf("}\n");
}

void	dump_desc(void)
{
	int index = 0;
	int r;

	printf(CMD_FMT " {\n", CMD_ARG(num_cmd));
	do {
	    r = handle_data(NULL, NULL,
		num_cmd, index, -3);
	    index++;
	} while ((r >= 0));
	printf("}\n");
}

void	do_syscall(union cmd_arg *data)
{
	num_ret = vserver(num_cmd, num_id, data);
}

struct err_data *get_err_data(int error)
{
	struct err_data *ptr;

	for (ptr=err_data; ptr->code; ptr++)
	    if (ptr->code == error)
		break;

	return ptr;
}


#define OPTIONS "hdi:klqtABC:V"

int	main(int argc, char *argv[])
{
	extern int optind;
	extern char *optarg;
	char c, errflg = 0;
	char dataflg;
	int r, index;
	union cmd_arg vc_data;

	cmd_name = argv[0];
	while ((c = getopt(argc, argv, "+" OPTIONS)) != EOF) {
	    switch (c) {
	    case 'h':
		fprintf(stderr,
		    "This is %s " VERSION "\n"
		    "options are:\n"
		    "-h        print this help message\n"
		    "-A        dump data after syscall\n"
		    "-B        dump data before syscall\n"
		    "-d        describe command data\n"
		    "-i <num>  id value\n"
		    "-k        continue after error\n"
		    "-l        list command and args\n"
		    "-q        quiet, returncode only\n"
		    "-t        terse, show only relevant\n"
		    "-C <cmd>  vcmd command name\n"
		    "-V        verify interface version\n"
		    "--        end of options\n"
		    ,cmd_name);
		exit(0);
		break;
	    case 'd':
		opt_desc = 1;
		break;
	    case 'i':
		num_id = ato_uint32_t(optarg);
		opt_id = 1;
		break;
	    case 'k':
		opt_cont = 1;
		break;
	    case 'l':
		opt_list = 1;
		break;
	    case 'q':
		opt_quiet = 1;
		break;
	    case 't':
		opt_terse = 1;
		break;
	    case 'A':
		opt_after = 1;
		break;
	    case 'B':
		opt_before = 1;
		break;
	    case 'C':
		num_cmd = parse_cmd(optarg);
		if (num_cmd == -1) {
		    printf("unrecognized vc %s, aborting.\n",
			optarg);
		    exit(3);
		}
		opt_cmd = 1;
		break;
	    case 'V':
		opt_vers = 1;
		break;
	    case '?':
	    default:
		errflg++;
		break;
	    }
	}
	if (errflg) {
	    fprintf(stderr, 
		"Usage: %s [-" OPTIONS "] <args> [-e cmd ...]\n"
		"%s -h for help.\n",
		cmd_name, cmd_name);
	    exit(2);
	}

	if (opt_vers) {
	    r = vserver(VCMD_get_version, 0, NULL);
	    if (r<0)
		perror("vc_get_version");
	    else 
		printf("version: %04x:%04x\n", 
		    (r>>16) & 0xFFFF, r & 0xFFFF);
	}

	memset(&vc_data, 0, sizeof(vc_data));
	dataflg = has_data(num_cmd);

	if (opt_desc) {
	    if (!dataflg)
		printf(CMD_FMT " does not take data.\n",
		    CMD_ARG(num_cmd));
	    else
		dump_desc();
	}

	if (!dataflg)
	    goto skip_data;

	index = 0;
	for (; optind < argc; optind++) {
	    if (strcmp(argv[optind], "--") == 0) {
		optind++;
		break;
	    }

	    /* skip to next input field */
	    while (!(get_io(num_cmd, index) & 1))
		index++;

	    /* handle named data */
	    if (argv[optind][0] == '.')
		index = set_named(&vc_data,
		    num_cmd, argv[optind]+1);
	    else
		index = set_index(&vc_data,
		    num_cmd, index, argv[optind]);
	    if (index == -1)
		printf("warning: argument %d %s ignored.\n",
		    optind, argv[optind]);
	}


	if (opt_before)
	    dump_data(&vc_data, 1);

skip_data:
	do_syscall(&vc_data);

	if (num_ret == -1) {
	    num_err = errno;
	    if (opt_desc) {
		struct err_data *ed = get_err_data(num_err);
		
		printf("vserver: %s [%d] %s\n",
		    ed->symb, num_err, ed->desc);
	    }
	    else if (!opt_quiet)
		perror("vserver");
	    if (!opt_cont)
		exit(-num_err);
	} else if (!opt_quiet) {
	    printf("vserver: ret = " FMT_int32_t "\n",
		(long long)num_ret, (long long)num_ret);
	}

	if (opt_after && dataflg)
	    dump_data(&vc_data, 2);

	if (opt_list) {
	    for (index=0; optind < argc; optind++, index++)
		printf("cmd[%d] = %s\n", index, argv[optind]);
	    optind -= index;
	}
	if (optind < argc) {
	    int ret;
	    
	    exec_name = argv[optind];
	    ret = execvp(exec_name, argv+optind);
	    
	    if (ret<0)
		perror("execvp");
	    exit(5);
	}

	exit(num_ret);
}

